@@ -142,7 +142,11 @@ def _merge_index(
     tmp_output_folder: Optional[str] = None,
     index_optimizer: Callable = None,
 ) -> Tuple[faiss.Index, Dict[str, str]]:
-    """Merge all the indices in `small_indices_folder` into single one return the merged index."""
+    """
+    Merge all the indices in `small_indices_folder` into single one.
+    Also run optimization when `index_optimizer` is given.
+    Returns the merged index and the metric
+    """
     fs = _get_file_system(small_indices_folder)
     small_indices_files = sorted(fs.ls(small_indices_folder, detail=False))
     small_indices_files = small_indices_files[start:end]
@@ -172,7 +176,7 @@ def _merge_index(
     first_index_file = small_indices_files[0]
     first_index_size = fs.size(first_index_file)
     max_sizes_in_bytes = cast_memory_to_bytes(max_size_on_disk)
-    nb_files_each_time = max(int(max_sizes_in_bytes / first_index_size), 1)
+    nb_files_each_time = max(1, int(max_sizes_in_bytes / first_index_size))
     merged_index = None
     n = len(small_indices_files)
     nb_iterations = max(math.ceil(n / nb_files_each_time), 1)
@@ -205,10 +209,15 @@ def _get_file_system(path: str) -> fsspec.AbstractFileSystem:
 
 
 def _merge_to_n_indices(spark_session, n: int, src_folder: str, dst_folder: str, index_optimizer: Callable = None):
-    """Merge all the indices from src_folder into n indices, and return the folder for the next stage"""
+    """Merge all the indices from src_folder into n indices, and return the folder for the next stage, as well as the metrics"""
     fs = _get_file_system(src_folder)
     nb_indices_on_src_folder = len(fs.ls(src_folder, detail=False))
-    batch_size = max(int(nb_indices_on_src_folder / n), 1)
+
+    if nb_indices_on_src_folder <= n and index_optimizer is None:
+        # no need to merge
+        return src_folder, None
+
+    batch_size = max(1, int(nb_indices_on_src_folder / n))
     n = math.ceil(nb_indices_on_src_folder / batch_size)
     merge_batches = _batch_loader(batch_size=batch_size, nb_batches=n)
 
@@ -232,7 +241,6 @@ def _merge_to_n_indices(spark_session, n: int, src_folder: str, dst_folder: str,
         metrics_dict = {metric_info["index_path"]: metric_info for metric_info in metrics}  # type: ignore
     else:
         metrics_dict = None  # type: ignore
-    fs = _get_file_system(src_folder)
     for file in fs.ls(src_folder, detail=False):
         if fs.isfile(file):
             fs.rm(file)
@@ -273,44 +281,44 @@ def run(
     """
     temporary_indices_folder = make_path_absolute(temporary_indices_folder)
     fs = _get_file_system(temporary_indices_folder)
+    if fs.exists(temporary_indices_folder):
+        fs.rm(temporary_indices_folder, recursive=True)
     stage1_folder = temporary_indices_folder.rstrip("/") + "/stage-1"
     ss = _get_pyspark_active_session()
-    if False:
-        # broadcast the index bytes
-        trained_index_bytes = get_bytes_from_index(faiss_index)
-        broadcast_trained_index_bytes = ss.sparkContext.broadcast(trained_index_bytes)
-        sc = ss._jsc.sc()  # pylint: disable=protected-access
-        n_workers = len(sc.statusTracker().getExecutorInfos()) - 1
+    # broadcast the index bytes
+    trained_index_bytes = get_bytes_from_index(faiss_index)
+    broadcast_trained_index_bytes = ss.sparkContext.broadcast(trained_index_bytes)
+    sc = ss._jsc.sc()  # pylint: disable=protected-access
+    n_workers = len(sc.statusTracker().getExecutorInfos()) - 1
 
-        # maximum between the number of spark workers, 100M embeddings per task and the number of indices to keep
-        estimated_nb_batches = max(n_workers, int(embedding_reader.count / (10 ** 7)), nb_indices_to_keep)
-        batch_size = max(int(embedding_reader.count / estimated_nb_batches), 1)
-        nb_batches = math.ceil(embedding_reader.count / batch_size)
+    # maximum between the number of spark workers, 100M embeddings per task and the number of indices to keep
+    estimated_nb_batches = max(n_workers, int(embedding_reader.count / (10 ** 7)), nb_indices_to_keep)
+    batch_size = max(1, int(embedding_reader.count / estimated_nb_batches))
+    nb_batches = math.ceil(embedding_reader.count / batch_size)
 
-        batches = _batch_loader(batch_size=batch_size, nb_batches=nb_batches)
-        rdd = ss.sparkContext.parallelize(batches, nb_batches)
-        with Timeit("-> Adding indices", indent=2):
-            rdd.foreach(
-                lambda x: _add_index(
-                    batch_id=x[0],
-                    start=x[1],
-                    end=x[2],
-                    memory_available_for_adding=memory_available_for_adding,
-                    broadcast_trained_index_bytes=broadcast_trained_index_bytes,
-                    embedding_reader=embedding_reader,
-                    small_indices_folder=stage1_folder,
-                    num_cores=num_cores_per_executor,
-                    embedding_ids_df_handler=embedding_ids_df_handler,
-                    nb_batches=nb_batches,
-                )
+    batches = _batch_loader(batch_size=batch_size, nb_batches=nb_batches)
+    rdd = ss.sparkContext.parallelize(batches, nb_batches)
+    with Timeit("-> Adding indices", indent=2):
+        rdd.foreach(
+            lambda x: _add_index(
+                batch_id=x[0],
+                start=x[1],
+                end=x[2],
+                memory_available_for_adding=memory_available_for_adding,
+                broadcast_trained_index_bytes=broadcast_trained_index_bytes,
+                embedding_reader=embedding_reader,
+                small_indices_folder=stage1_folder,
+                num_cores=num_cores_per_executor,
+                embedding_ids_df_handler=embedding_ids_df_handler,
+                nb_batches=nb_batches,
             )
+        )
 
     with Timeit("-> Merging indices", indent=2):
         stage2_folder = temporary_indices_folder.rstrip("/") + "/stage-2"
-        #next_stage_folder, _ = _merge_to_n_indices(
-        #    spark_session=ss, n=100, src_folder=stage1_folder, dst_folder=stage2_folder, index_optimizer=None
-        #)
-        next_stage_folder = stage2_folder
+        next_stage_folder, _ = _merge_to_n_indices(
+            spark_session=ss, n=100, src_folder=stage1_folder, dst_folder=stage2_folder, index_optimizer=None
+        )
         if nb_indices_to_keep == 1:
             merged_index, _ = _merge_index(small_indices_folder=next_stage_folder, nb_batches=1)
             fs.rm(temporary_indices_folder, recursive=True)
@@ -325,4 +333,5 @@ def run(
                 dst_folder=final_folder,
                 index_optimizer=index_optimizer,
             )
+            fs.rm(temporary_indices_folder, recursive=True)
             return None, metrics
